//  DtL - Derive to LaTeX by Gary Bishop 29 June 1999

#include <stdio.h>
#include <windows.h>
#include <ctype.h>
#include "Str.h"
#include "DtL.h"

static char rcsid[] = "$Id: io.cpp,v 1.1 1999-07-01 11:03:08-04 gb Exp gb $";

extern FILE* yyin;

int DtL_read_clipboard(char* obuf, int max_size)
{
  static int clipboard_fmt = 0;
  static Str buff;

  if(clipboard_fmt == 0) {
    clipboard_fmt = RegisterClipboardFormat("DfW Expression");
    if(clipboard_fmt == 0) {
      fprintf(stderr, "RegisterClipboardFormat failed\n");
      exit(1);
    }
    if(OpenClipboard(NULL) == 0) {
      fprintf(stderr, "OpenClipboard failed\n");
      exit(1);
    }
    HANDLE h = GetClipboardData(clipboard_fmt);
    if(h == NULL) {
      fprintf(stderr, "GetClipboardData failed\n");
      exit(1);
    }
    char* lptstr = (char*)GlobalLock(h);
    if (lptstr == NULL) {
      fprintf(stderr, "GlobalLock Failed\n");
      exit(1);
    }
    buff = Str(lptstr);
    GlobalUnlock(h);
    CloseClipboard();
  }
  static int ndx = 0;
  if(ndx < buff.length()) {
    obuf[0] = buff[ndx++];
    return 1;
  } else
    return 0;
}

// translate characters read from a file into the character set used internally

static char translate_tab[256];

static void init_translate_tab()
{
  char* tt = translate_tab;
  for(int i=0; i<256; i++)
    tt[i] = i;
  tt[0233]='\242';
  tt[0246]='\252';
  tt[0247]='\272';
  tt[0250]='\277';
  tt[0253]='\275';
  tt[0255]='\241';
  tt[0271]='\235';
  tt[0272]='\236';
  tt[0274]='\327';
  tt[0275]='\256';
  tt[0276]='\263';
  tt[0301]='\264';
  tt[0303]='\270';
  tt[0305]='\276';
  tt[0321]='\320';
  tt[0327]='\247';
  tt[0340]='\251';
  tt[0341]='\337';
  tt[0345]='\336';
  tt[0346]='\265';
  tt[0347]='\274';
  tt[0350]='\245';
  tt[0351]='\262';
  tt[0352]='\250';
  tt[0353]='\253';
  tt[0356]='\255';
  tt[0367]='\367';
  tt[0374]='\257';
  tt[0376]='\376';
}


int DtL_read_file(char* buf, int max_size)
{
  int c = getc(yyin);
  if(c == EOF)
    return 0;
  else if(c == '~') {
    int c2 = getc(yyin);
    if(c2 == '\r') {
      int c3 = getc(yyin);
      if(c3 == '\n') {
	lineno++;
	return DtL_read_file(buf, max_size);
      } else if(c3 == EOF)
	return 0;
      else {
	buf[0] = translate_tab[c3];
	return 1;
      }
    } else if(c2 == '\n') {
      lineno++;
      return DtL_read_file(buf, max_size);
    } else {
      buf[0] = translate_tab[c2];
      return 1;
    }
  } else {
    buf[0] = translate_tab[c];
    return 1;
  }
}

int DtL_input(char* buf, int max_size)
{
  int r=0;
  switch(DataRead) {
  case DP_Clipboard:
    r=DtL_read_clipboard(buf, max_size);
    break;
  case DP_File:
    r=DtL_read_file(buf, max_size);
    break;
  }
  if(yydebug && r != 0)
    fprintf(stderr, "input = '%c':%o\n", buf[0], buf[0]&0xff);
  return r;
}

void DtL_write_clipboard(const Str& s)
{
  HGLOBAL h = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, s.length()+1);
  if(h == NULL) {
    fprintf(stderr, "GlobalAlloc failed\n");
    exit(1);
  }
  char* p = (char*)GlobalLock(h);
  if(p == 0) {
    fprintf(stderr, "GlobalLock failed\n");
    exit(1);
  }
  strcpy(p, (const char*)s);
  GlobalUnlock(h);

  if(OpenClipboard(NULL) == 0) {
    fprintf(stderr, "OpenClipboard failed\n");
    exit(1);
  }
  if(SetClipboardData(CF_TEXT, h) == NULL) {
    fprintf(stderr, "SetClipboardData failed\n");
    exit(1);
  }
  CloseClipboard();
}

void DtL_write_file(const Str& s)
{
  puts((const char*)s);
  fflush(stdout);
}

void DtL_output(const Str& comment, const Str& s)
{
  Str out = s;
  if(KeepComments)
    out = "%" + comment + "\n" + s;

  switch(DataWrite) {
  case DP_Clipboard:
    DtL_write_clipboard(out);
    break;
  case DP_File:
    DtL_write_file(out);
    break;
  }
}

struct {
  char* src;
  char* latex;
} builtins[] = {
  { "alpha", "\\alpha" },
  { "beta", "\\beta" },
  { "gamma", "\\gamma" },
  { "GAMMA", "\\Gamma" },
  { "delta", "\\delta" },
  { "DELTA", "\\Delta" },
  { "epsilon", "\\epsilon" },
  { "zeta", "\\zeta" },
  { "eta", "\\eta" },
  { "theta", "\\theta" },
  { "THETA", "\\Theta" },
  { "iota", "\\iota" },
  { "kappa", "\\kappa" },
  { "lambda", "\\lambda" },
  { "LAMBDA", "\\Lambda" },
  { "mu", "\\mu" },
  { "nu", "\\nu" },
  { "xi", "\\xi" },
  { "XI", "\\Xi" },
  { "pi", "\\pi" },
  { "PI", "\\Pi" },
  { "rho", "\\rho" },
  { "sigma", "\\sigma" },
  { "SIGMA", "\\Sigma" },
  { "tau", "\\tau" },
  { "upsilon", "\\upsilon" },
  { "UPSILON", "\\Upsilon" },
  { "phi", "\\phi" },
  { "PHI", "\\Phi" },
  { "chi", "\\chi" },
  { "psi", "\\psi" },
  { "PSI", "\\Psi" },
  { "omega", "\\omega" },
  { "OMEGA", "\\Omega" },
  { "SIN", "\\sin" },
  { "COS", "\\cos" },
  { "TAN", "\\tan" },
  { "COT", "\\cot" },
  { "SEC", "\\sec" },
  { "CSC", "\\csc" },
  { "ATAN", "\\arctan" },
  { "ACOT", "\\arccot" },
  { "ASIN", "\\arcsin" },
  { "ACOS", "\\arccos" },
  { "ASEC", "\\mathrm{arcsec}" },
  { "ACSC", "\\mathrm{arccsc}" },
  { "EXP", "\\exp" },
  { "LOG", "\\log" },
  { "LN", "\\ln" },
  { "UNION", "\\cup" },
  { "SUBSET", "\\subseteq" },
  { "PRODUCT", "\\Pi" },
  { "SUM", "\\Sigma" },
  { "INTERSECTION", "\\cap" },
  { "NOT", "\\neg" },
  { "deg", "\\deg" },
  { "SQRT", "\\sqrt" },
  { "SIGN", "\\mathrm{sign}" },
  { "inf", "\\infty" },
  { "DET", "\\det" },
};

const int builtins_size = sizeof(builtins)/sizeof(builtins[0]);
char* builtin(const Str& s)
{
  for(int i=0; i<builtins_size; i++)
    if(strcmp(s, builtins[i].src) == 0)
      return builtins[i].latex;
  return 0;
}

char* special_char[256];

void init_special_char()
{
  for(int i=0; i<256; i++)
    special_char[i] = 0;

  char** sc = special_char;
  sc[0213]="\\sqrt";
  sc[0215]="\\cap";
  sc[0216]="\\cup";
  sc[0217]="\\wedge";
  sc[0220]="\\vee";
  sc[0221]="\\subseteq";
  sc[0222]="\\geq";
  sc[0223]="\\leq";
  sc[0224]="\\neq";
  sc[0267]="\\cdot";
  sc[0226]="\\infty";
  sc[0233]="\\Gamma";
  sc[0235]="\\Delta";
  sc[0236]="\\Theta";
  sc[0241]="\\Lambda";
  sc[0242]="\\Xi";
  sc[0243]="\\Pi";
  sc[0244]="\\Sigma";
  sc[0245]="\\Phi";
  sc[0247]="\\psi";
  sc[0250]="\\Omega";
  sc[0251]="\\alpha";
  sc[0252]="\\gamma";
  sc[0253]="\\delta";
  sc[0254]="\\neg";
  sc[0255]="\\epsilon";
  sc[0256]="\\zeta";
  sc[0257]="\\eta";
  sc[0260]="\\deg";
  sc[0261]="\\pm";
  sc[0262]="\\theta";
  sc[0263]="\\iota";
  sc[0264]="\\kappa";
  sc[0265]="\\mu";
  sc[0270]="\\xi";
  sc[0271]="\\pi";
  sc[0272]="\\rho";
  sc[0274]="\\tau";
  sc[0275]="\\upsilon";
  sc[0276]="\\phi";
  sc[0277]="\\lambda";
  sc[0320]="\\Upsilon";
  sc[0327]="\\Psi";
  sc[0336]="\\sigma";
  sc[0337]="\\beta";
  sc[0352]="\\mathbf{e}";
  sc[0356]="\\mathbf{I}";
  sc[0367]="\\omega";
  sc[0376]="\\nu";
}  

Str mathit(const Str& s)
{
  char* sym = 0;
  // the idea here is to prepare a string to serve as a variable name.
  if(!isallascii(s)) { // if it is not ascii then it might be a coded builtin
    if(s.length() == 1) { // a single special character
      if((sym = special_char[s[0] & 0xff])) {
	return sym;
      } else {
	fprintf(stderr, "unexpected non-ascii character ignored in mathit\n");
	return "?";
      }
    } else { // more complicated case of multiple characters, some of which may be special
      Str result = "\\mathit{";
      for(int i=0; i<s.length(); i++) {
	int c = s[i] & 0xff;
	if(c >= 128) { // it's special
	  if((sym = special_char[c]))
	    result += sym[0]=='\\' ? Str(sym) + " " : Str(sym);
	  else {
	    fprintf(stderr, "unexpected non-ascii character ignored in mathit\n");
	    result += "?";
	  }
	} else { // it is a plain character
	  result += (char)c;
	}
      }
      return result + "}";
    }
  } else if(s.length() == 1 && isalpha(s[0])) {
    return s;
  } else if(isalllower(s) || isallupper(s)) { // single case ids might be builtin keywords
    if((sym = builtin(s))) {
       return sym;
    } else {
      return "\\mathit{" + s + "}";
    }
  } else { // must just be a mixed case name
    return "\\mathit{" + s + "}";
  }
}

Str mathrm(const Str& s)
{
  char* sym = 0;
  // the idea here is to prepare a string to serve as a variable name.
  if(!isallascii(s)) { // if it is not ascii then it might be a coded builtin
    if(s.length() == 1) { // a single special character
      if((sym = special_char[s[0] & 0xff])) {
	return sym;
      } else {
	fprintf(stderr, "unexpected non-ascii character ignored in mathit\n");
	return "?";
      }
    } else { // more complicated case of multiple characters, some of which may be special
      Str result = "\\mathrm{";
      for(int i=0; i<s.length(); i++) {
	int c = s[i] & 0xff;
	if(c >= 128) { // it's special
	  if((sym = special_char[c]))
	    result += sym[0]=='\\' ? Str(sym) + " " : Str(sym);
	  else {
	    fprintf(stderr, "unexpected non-ascii character ignored in mathit\n");
	    result += "?";
	  }
	} else { // it is a plain character
	  result += (char)c;
	}
      }
      return result + "}";
    }
  } else if(s.length() == 1 && isalpha(s[0])) {
    return s;
  } else if(isalllower(s) || isallupper(s)) { // single case ids might be builtin keywords
    if((sym = builtin(s))) {
       return sym;
    } else {
      return "\\mathrm{" + s + "}";
    }
  } else { // must just be a mixed case name
    return "\\mathrm{" + s + "}";
  }
}

void init_io()
{
  init_translate_tab();
  init_special_char();
}
